/*
//	@(#) getentropy.c - read some bytes from /dev/random or
//		fallback to rand48 (quick and dirty.)
//
//	License: Creative Commons CC0 
//		http://creativecommons.org/publicdomain/zero/1.0/legalcode
//	Author:	James Sainsbury
//		toves@sdf.lonestar.org
//
*/
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/select.h>

# include	"entropy.h"
static	int	rand48_getentropy (char entropy[], size_t entropy_size);
static	int	sys_getentropy (int fd, char entropy[], size_t entropy_size);

# if	!defined (_PATH_DEVRANDOM)
# define	DEVRANDOM	"/dev/random"
# else
# define	DEVRANDOM	_PATH_DEVRANDOM
# endif
static	const	char	rand_dev[]	= DEVRANDOM;

static	int	sys_getentropy (int fd, char entropy[], size_t entropy_size) {
	int	result	= -1;
	struct	timeval	timeout	= { .tv_sec = 5, .tv_usec = 0 };
	fd_set	inp;
	ssize_t	left	= entropy_size;
	ssize_t	n	= 0;
	FD_ZERO (&inp);
	FD_SET (fd, &inp);
	do	{
		int	nfds	= select (fd+1, &inp, 0, 0, &timeout);
		if (nfds == 1 && FD_ISSET (fd, &inp)) {
			n	= read (fd, entropy, left);
			left	-= n;
		}
		else if (nfds <= 0 || n == 0) {	// Just bail to software
			rand48_getentropy (&entropy[entropy_size-left], left);
			left	= 0;
		}
	} while (n > 0 && left > 0);
	result	= 0;
	return	result;
}

static	int	rand48_getentropy (char entropy[], size_t entropy_size) {
	unsigned short	seed[3];
	int	i	= 0;
	int	step	= sizeof(RAND_MAX); // Xrand48 return 32 bit "longs"	
	long	r	= jrand48(seed);
	while ((i+step) < entropy_size) {
		memcpy (&entropy[i], &r, step);
		long	r	= jrand48(seed);
		i	+= step;
	}
	memcpy (&entropy[i], &r, (entropy_size - i));
	return	0;
}

int	getentropy (char entropy[], size_t entropy_size) {
	int	result	= -1;
	int	fd	= open (rand_dev, O_RDONLY);
	if (fd>=0) {
		result	= sys_getentropy (fd, entropy, entropy_size);
		close (fd);
	}
	else	{
		result	= rand48_getentropy (entropy, entropy_size);
	}
	return	result;
}
