/*
//	@(#) genbfpw.c - generate a blowfish password
//
//	License: Creative Commons CC0 
//		http://creativecommons.org/publicdomain/zero/1.0/legalcode
//	Author:	James Sainsbury
//		toves@sdf.lonestar.org
//
*/
#include <unistd.h>
#include <crypt.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/select.h>

# include	"entropy.h"
/*
//	SuSE doesn't have the header file for the OW routines
*/
# include	"missing.h"

// r -> reentrant, n -> not malloc'd cf _ra
// crypt_gensalt_rn (prefix -> specify crypto $1 -> md5, $2a -> bf
//		iteration_count
//		entropy
//		entropy_size
//		output	-> the generated salt
//		output_size) returns &output[0]

const	char	CX_BLOWFISH[]	= "$2a$";
enum	{
	BF_BYTE_BITS		= 8,
	BF_ENTROPY_BITS		= 128,
	BF_ENTROPY_BYTES	= (BF_ENTROPY_BITS/BF_BYTE_BITS),
	BF_PREFIX_LEN		= 7,	// "$2a$10$"
	BF_SALT_LEN		= 22,
	BF_SETTING_LEN		= (BF_PREFIX_LEN+BF_SALT_LEN),
	BF_HASH_LEN		= 31,
	BF_TOTAL_LEN		= (BF_SETTING_LEN+BF_HASH_LEN),
	BF_ROUNDS		= 10,	// log2 Nrounds
	
};

static	char*	progname;
void	Usage(void) {
	fprintf (stderr, "Usage: %s [-s setting_string] [-p password_clear]\n");
	exit (EXIT_FAILURE);
}
void	fatal(const char* msg) {
	fprintf (stderr, "%s: Fatal: %s\n", progname, msg);
	exit (EXIT_FAILURE);
}
main (int argc, char* argv[]) {
	char	entropy [BF_ENTROPY_BYTES];
	char	settings [BF_SETTING_LEN+1];
	char	hashed_password [BF_TOTAL_LEN+1];
	char	clear_text [BUFSIZ];
	int	opt	= EOF;
	int	s_flag	= 0;
	int	p_flag	= 0;
	char*	r	= 0;
	progname	= argv[0];
	
	while ((opt = getopt (argc, argv, "s:p:")) != EOF) {
		switch (opt) {
		case	's': {
			size_t	len	= strlen (optarg);
			++s_flag;
			if(len >= BF_SETTING_LEN) {
				memcpy (settings, optarg, BF_SETTING_LEN);
				settings [BF_SETTING_LEN+1]	= '\0';
			}
			else	fatal ("settings too short\n");
		}
		break;
		case	'p': {
			size_t	len	= strlen (optarg);	
			++p_flag;
			if (len > 0 && len < BUFSIZ) {
				memcpy (clear_text, optarg, len);
				clear_text[len]	= '\0';
			}
			else	fatal ("clear text password too big/small\n");
		}
		break;
		default:
			Usage();
		break;
		}
	}
	
	if (!s_flag) {
		getentropy (entropy, sizeof(entropy));
		
		r	= crypt_gensalt_rn (CX_BLOWFISH, BF_ROUNDS,
				entropy, BF_ENTROPY_BYTES,
				settings, BF_SETTING_LEN+1);
		if (!r) {
			fatal (strerror(errno));
		}
	}
	if (!p_flag) { // Normally wouldn't echo but we are just playing
		char*	t	= clear_text;
		fputs ("? ", stdout);
		fgets (clear_text, BUFSIZ, stdin);
		t	= strchr(t, '\n');
		if (t) { *t	= '\0';}
	}
	// Needs to be cleared before use
	memset (hashed_password, 0, sizeof(hashed_password));
	r	= crypt_rn (clear_text, settings,
			hashed_password, BF_TOTAL_LEN+1);

	// The compiler can optimize this away 
	memset (clear_text, 0, sizeof(clear_text));

	if (!r) {
		fatal (strerror(errno));
	}

	fputs (hashed_password, stdout);
	fputc ('\n', stdout);

	exit (EXIT_SUCCESS);
}

